import argcomplete
import argparse
import collections
import collections.abc
import copy
import enum
import functools
import importlib
import importlib.util
import inspect
import os
import pkgutil
import re
import sys

import plumbum
import plumbum.commands
from plumbum import colors

from descriptors import (
    cachedproperty as _cachedproperty,
    classproperty as _classproperty,
)


__version__ = '1.1.0'

__all__ = (
    'cmd',
    'cmdmethod',
    'local',
    'localmethod',
    'main',
    'entrypoint',
    'init_package',
    'Command',
    'RootCommand',
    'Local',
    'LocalRoot',
)


def _friendly_version(version_info):
    return '.'.join(map(str, version_info[:3]))


def check_version(minimum_version, version_info=sys.version_info):
    if version_info < minimum_version:
        raise EnvironmentError(
            "{self} requires Python version {} or higher, not: {}"
            .format(
                *(_friendly_version(info) for info in (minimum_version,
                                                       version_info)),
                self=sys.argv[0].strip('./'),
            )
        )


def _noop(*args, **kwargs):
    pass


def main(command_class,
         minimum_version=(0,),
         argv=None,
         outfile=sys.stdout,
         extend_parser=_noop):
    args = None
    try:
        check_version(minimum_version)
        (parser, args) = command_class._get_parser_()
        extend_parser(parser)
        argcomplete.autocomplete(parser)
        parser.parse_args(argv, args)
        command = args._command_
        command.call()
    except Exception as exc:
        if args is None or getattr(args, 'traceback', True):
            raise

        print(f"[{exc.__class__.__name__}]" | colors.yellow,
              str(exc) | colors.red,
              file=outfile)
        sys.exit(1)
    except KeyboardInterrupt:
        print("stopped", file=outfile)


def add_manage_file(parser):
    parser.add_argument(
        '--manage-file',
        metavar='PATH',
        help="path to a manage command file",
    )


MANAGE_FILE_PATHS = (
    os.path.join('manage', '__init__.py'),
    'manage.py',
)


def init_package(path=None, name='manage'):
    """Initialize (import) the submodules, and recursively the
    subpackages, of a "manage" package at ``path``.

    ``path`` may be specified as either a system directory path or a
    list of these.

    If ``path`` is unspecified, it is inferred from the already-imported
    "manage" top-level module.

    """
    if path is None:
        manager = sys.modules[name]
        init_package(manager.__path__, name)
        return

    if isinstance(path, str):
        init_package([path], name)
        return

    for module_info in pkgutil.walk_packages(path, f'{name}.'):
        if not module_info.ispkg:
            importlib.import_module(module_info.name)


def execute(argv=None):
    parser = argparse.ArgumentParser(add_help=False)
    add_manage_file(parser)
    (args, _remainder) = parser.parse_known_args(argv)

    # import manage.py / manage package
    manager = None

    # 1) NBD, already on the python-path
    if not args.manage_file:
        try:
            manager = importlib.import_module('manage')
        except ImportError:
            pass

    if not manager:
        # CWD may not be in PYTHONPATH (and that's OK);
        # or, may have specified alternative location
        if args.manage_file:
            # 2) alternative location specified
            if os.path.isdir(args.manage_file):
                file_path = os.path.join(args.manage_file, '__init__.py')
            else:
                file_path = args.manage_file
        else:
            # 3) we'll try the usual suspects
            # Note: *could* raise the error here, in a for/else block
            for file_path in MANAGE_FILE_PATHS:
                if os.path.isfile(file_path):
                    break

        spec = importlib.util.spec_from_file_location('manage', file_path)
        manager = importlib.util.module_from_spec(spec)

        # in case we're loading a package (with internal references)
        sys.modules['manage'] = manager

        # prevent __pycache__ clutter
        dont_write_bytecode = sys.dont_write_bytecode
        sys.dont_write_bytecode = True

        try:
            spec.loader.exec_module(manager)
        except FileNotFoundError:
            print("could not import module 'manage' and file not found:" | colors.red,
                  file_path,
                  end='\n\n')
            with colors.cyan:
                print("hint: specify a manage file path via --manage-file, e.g.:",
                      '\tmanage --manage-file=path/to/project/manage-file.py ...',
                      "...or set PYTHONPATH to the directory containing your manage.py, e.g.:",
                      '\tPYTHONPATH=path/to/project manage ...',
                      sep='\n\n')
            sys.exit(1)
        finally:
            sys.dont_write_bytecode = dont_write_bytecode

    # auto-init manage package's submodules and subpackages
    auto_init_package = getattr(manager, '__auto_init_package__', True)
    submodule_search_locations = getattr(manager, '__path__', None)
    if auto_init_package and submodule_search_locations is not None:
        init_package(submodule_search_locations)

    # determine entrypoint
    for command_filter in (
        _is_entrypoint_,
        _is_root_command_,
        _is_command_,
    ):
        candidates = (value for value in vars(manager).values()
                      if command_filter(value))

        try:
            entrypoint = next(candidates)
        except StopIteration:
            # no matches for this filter. try the next one.
            continue

        try:
            next(candidates)
        except StopIteration:
            # only one match. run it!
            main(entrypoint, argv=argv, extend_parser=add_manage_file)
            return

        print("multiple entrypoints found" | colors.red,
              end='\n\n')
        with colors.cyan:
            print("hint: define one root command or mark "
                  "an entrypoint with the entrypoint decorator")
        sys.exit(1)

    print("no entrypoint found. define at least one command." | colors.yellow)
    sys.exit(1)


def _is_entrypoint_(obj):
    return getattr(obj, '_argcmdr_entrypoint_', False)


def _is_root_command_(obj):
    return (isinstance(obj, type) and
            issubclass(obj, RootCommand) and
            obj is not RootCommand and
            obj is not LocalRoot)


def _is_command_(obj):
    return (isinstance(obj, type) and
            issubclass(obj, Command) and
            obj is not Command and
            obj is not Local)


_exhaust_iterable = collections.deque(maxlen=0).extend

_command_lookup_error_message = ("command hierarchy indices must be str (to descend), or "
                                 "negative integer (to ascend), not %r")


class Command:

    #
    # interface *intended* for override/extension by subclasses
    #

    formatter_class = argparse.HelpFormatter

    @_classproperty
    def help(cls):
        if cls.__doc__ is None:
            return None

        parts = re.split(r'(?:\. )|\n', cls.__doc__, 1)
        return parts[0]

    @_classproperty
    def name(cls):
        return cls.__name__.lower()

    def __init__(self, parser):
        self._args_ = None
        self._parser_ = parser

        self._children_ = None
        self._parents_ = None

    def __call__(self, args):
        args._parser_.print_usage()

    #
    # public interface
    #

    def __getitem__(self, key):
        if isinstance(key, (str, bytes)) or not isinstance(key, collections.abc.Sequence):
            return self.__getitem__((key,))

        if not key:
            return self

        head = key[0]

        if isinstance(head, str):
            item = self._get_children_().get(head)

            if item is None:
                raise KeyError(f"command {self.name} has no child {head!r}")
        elif isinstance(head, int):
            if head >= 0:
                raise ValueError(_command_lookup_error_message % head)

            try:
                item = self._get_parents_()[-1 - head]
            except IndexError:
                raise IndexError(f"command {self.name} has no parent {head!r}")
        else:
            raise TypeError(_command_lookup_error_message % head)

        return item.__getitem__(key[1:])

    def __iter__(self):
        yield from self._get_children_().values()

    @property
    def root(self):
        try:
            return self._get_parents_()[-1]
        except IndexError:
            return None

    def get_args(self):
        args = getattr(self, '_args_', None)  # user *may* neglect to super().__init__

        if args is None:
            raise RuntimeError('parsed argument namespace not available at this stage')

        return args

    @property
    def args(self):
        args = self.get_args()

        if args._command_ is not self:
            # The property is here made consistent with the argumentation of
            # the delegate() interface; and, methods relying on this property
            # are thereby enabled regardless of whether their associated
            # command was invoked directly or via delegation.
            #
            # Note: It is arguable that this is unnecessary in the case that
            # the requesting command is an ancestor of the CLI-invoked
            # command -- in this case, all of the requesting command's
            # arguments should be filled, and there's little need.
            #
            # This condition could be tested here with the expression:
            #
            #     self (not) in args._command_._parents_
            #
            # (and in which case, for consistency, delegate() should perhaps be
            # modified to also pass non-delegate args to the ancestor command).
            #
            # However, more than just the CLI arguments, delegate_args further
            # populates the command's own sub-parser. And so, for now, for
            # consistency and relative simplicity, all commands will *always*
            # get their "own" args and parser, via delegate_args whenever it is
            # not the CLI-invoked command.
            #
            return self.delegate_args

        return args

    @_cachedproperty
    def delegate_args(self):
        args = copy.copy(self.get_args())
        args._parser_ = self._parser_

        for action in self._parser_._actions:
            args.__dict__.setdefault(action.dest, action.default)

        for default in self._parser_._defaults.items():
            args.__dict__.setdefault(*default)

        return args

    @property
    def parser(self):
        return getattr(self, '_parser_', None)  # user *may* neglect to super().__init__

    def call(self, *additional):
        return self.delegate('__call__', *additional)

    def delegate(self, method_name='__call__', *additional):
        nspace = self.args
        call_args = (nspace, nspace._parser_) + additional
        call_arg_count = len(call_args)

        target_callable = getattr(self, method_name)
        signature = inspect.signature(target_callable)
        parameters = [name for (index, (name, param)) in enumerate(signature.parameters.items())
                      if index < call_arg_count or param.default is param.empty]
        param_count = len(parameters)

        if param_count > call_arg_count:
            raise TypeError(
                f"{self.__class__.__name__}.{method_name}() "
                "requires too many positional arguments: " +
                ', '.join(repr(param) for param in parameters[call_arg_count:])
            )

        return target_callable(*call_args[:param_count])

    #
    # interface which *may* be overriden/extended by subclasses to customize operation
    #

    _allow_traceback_ = True

    @_classproperty
    def _subcommands_(cls):
        return [value for value in vars(cls).values()
                if isinstance(value, type) and issubclass(value, Command)]

    @staticmethod
    def _new_namespace_():
        return argparse.Namespace()

    @classmethod
    def _new_parser_(cls):
        parser = argparse.ArgumentParser(description=cls.__doc__,
                                         formatter_class=cls.formatter_class)

        if cls._allow_traceback_:
            parser.add_argument(
                '--tb', '--traceback',
                action='store_true',
                default=False,
                dest='traceback',
                help="print error tracebacks",
            )

        return parser

    @staticmethod
    def _new_subparser_(subparsers, subcommand):
        return subparsers.add_parser(subcommand.name,
                                     description=subcommand.__doc__,
                                     help=subcommand.help,
                                     formatter_class=subcommand.formatter_class)

    @classmethod
    def _build_interface_(cls, parser=None, namespace=None, chain=None, parents=()):
        if parser is None:
            parser = cls._new_parser_()
        if namespace is None:
            namespace = cls._new_namespace_()

        command = cls(parser)
        command._parents_ = parents
        command._children_ = {}
        command._args_ = namespace
        command._parser_ = parser  # user *may* neglect to super().__init__

        if chain is not None:
            chain[cls.name] = command

        parser.set_defaults(
            _command_=command,
            _parser_=parser,
        )
        yield (parser, namespace, command)

        subparsers = None
        subparents = (command,) + parents
        for subcommand in cls._subcommands_:
            if subparsers is None:
                subparsers = parser.add_subparsers(
                    title="{} commands".format(cls.name),
                    help="available commands",
                )

            subparser = cls._new_subparser_(subparsers, subcommand)

            yield from subcommand._build_interface_(subparser,
                                                    namespace,
                                                    command._children_,
                                                    subparents)

    @classmethod
    def _init_parser_(cls, parser, namespace):
        interface = cls._build_interface_(parser, namespace)
        _exhaust_iterable(interface)

    @classmethod
    def _get_parser_(cls):
        parser = cls._new_parser_()
        namespace = cls._new_namespace_()
        cls._init_parser_(parser, namespace)
        return (parser, namespace)

    #
    # internal helpers
    #

    def _get_children_(self):
        children = getattr(self, '_children_', None)  # user *may* neglect to super().__init__

        if children is None:
            raise RuntimeError('hierarchy of constructed commands not available at this stage')

        return children

    def _get_parents_(self):
        parents = getattr(self, '_parents_', None)  # user *may* neglect to super().__init__

        if parents is None:
            raise RuntimeError('hierarchy of constructed commands not available at this stage')

        return parents


class RootCommand(Command):

    _registry_ = None

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._registry_ = []

    @classmethod
    def register(cls, subcommand):
        if cls._registry_ is None:
            raise TypeError(f"{cls.__name__} is abstract and may not register subcommands")

        cls._registry_.append(subcommand)
        return subcommand

    @_classproperty
    def _subcommands_(cls):
        subcommands = super()._subcommands_
        if cls._registry_ is not None:
            subcommands += cls._registry_
        return subcommands


class CacheDict(collections.defaultdict):

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)

        value = self[key] = self.default_factory(key)
        return value


class _SHH(plumbum.commands.ExecutionModifier):
    """plumbum execution modifier to ensure output is not echoed to terminal

    essentially a no-op, this may be used to override argcmdr settings
    and cli flags controlling this feature, on a line-by-line basis, to
    hide unnecessary or problematic (e.g. highly verbose) command output.

    """
    __slots__ = ('retcode', 'timeout')

    def __init__(self, retcode=0, timeout=None):
        self.retcode = retcode
        self.timeout = timeout

    def __rand__(self, cmd):
        return cmd.run(retcode=self.retcode, timeout=self.timeout)


SHH = _SHH()


class Local(Command):

    local = CacheDict(plumbum.local.__getitem__)

    # link common exceptions
    local.ProcessExecutionError = plumbum.ProcessExecutionError
    local.CommandNotFound = plumbum.CommandNotFound

    # ...and modifiers
    local.BG = plumbum.BG
    local.FG = plumbum.FG
    local.TEE = plumbum.TEE

    # ...(as well as our own)
    local.SHH = SHH

    _run_kws_ = frozenset(('retcode', 'timeout'))

    @classmethod
    def _new_parser_(cls):
        parser = super()._new_parser_()
        parser.add_argument(
            '-q', '--quiet',
            action='store_false',
            default=True,
            dest='foreground',
            help="do not print command output",
        )
        parser.add_argument(
            '-d', '--dry-run',
            action='store_false',
            default=True,
            dest='execute_commands',
            help="do not execute commands, "
                 "but print what they are (unless --no-show is provided)",
        )
        parser.add_argument(
            '-s', '--show',
            action='store_true',
            default=None,
            dest='show_commands',
            help="print command expressions "
                 "(by default not printed unless dry-run)",
        )
        parser.add_argument(
            '--no-show',
            action='store_false',
            default=None,
            dest='show_commands',
            help="do not print command expressions "
                 "(by default not printed unless dry-run)",
        )
        return parser

    def _show_command_(self, command, force=False):
        if force or self.args.show_commands or (
            self.args.show_commands is None and
            not self.args.execute_commands
        ):
            # Due to plumbum issue #398 we can't do this:
            #
            # print('>', colors['#5FAF5F'] | str(command))
            #
            # whenever the command has been supplied stdin.
            formulation = ' '.join(map(str, command.formulate()))
            print('>', colors['#5FAF5F'] | formulation)

    def delegate(self, method_name='prepare', *additional):
        return super().delegate(method_name, self.local, *additional)

    def __call__(self, args):
        commands = self.delegate()

        if commands is None:
            return

        if isinstance(commands, plumbum.commands.BaseCommand) or (
            isinstance(commands, (tuple, list)) and
            len(commands) == 2 and
            isinstance(commands[0], plumbum.commands.ExecutionModifier)
        ):
            commands = (commands,)

        send = hasattr(commands, 'send')
        run_kwargs = {key: getattr(self.prepare, key)
                      for key in self._run_kws_ & self.prepare.__dict__.keys()}

        result = thrown = None
        empty_result = (None, None, None)
        iterator = iter(commands)

        while True:
            try:
                if send and result is not None:
                    command = iterator.send(result)
                elif send and thrown is not None:
                    command = iterator.throw(thrown)
                else:
                    command = next(iterator)
            except StopIteration:
                break

            if isinstance(command, (tuple, list)):
                (modifier, command) = command
            else:
                modifier = None

            self._show_command_(command)

            if args.execute_commands:
                try:
                    if modifier is not None:
                        if run_kwargs:
                            result = command & modifier(**run_kwargs)
                        else:
                            result = command & modifier

                        if result is None:
                            result = empty_result
                    elif args.foreground:
                        result = command & plumbum.TEE(**run_kwargs)
                    else:
                        result = command.run(**run_kwargs)
                except Exception as exc:
                    if not send:
                        raise

                    result = None
                    thrown = exc
                else:
                    thrown = None
            else:
                result = empty_result

    def prepare(self, args):
        super().__call__(args)


class LocalRoot(Local, RootCommand):
    pass


#                                   #
# command manufacture via decorator #
#                                   #

Unset = object()


class GeneratedCommand:
    """Mix-in for manufactured commands."""

    _parser_args_ = ()

    def __init__(self, parser):
        super().__init__(parser)

        for (args, kwargs) in self._parser_args_:
            parser.add_argument(*args, **kwargs)


class CommandMethod:
    """Decorating descriptor producing a command functionality method
    which receives as its first argument its parent command instance.

    """
    def __init__(self, func):
        self.__func__ = func

    def __get__(self, instance, cls=None):
        if instance is None:
            return self

        parent = instance[-1]
        return self.__func__.__get__(parent)


class WrappedCallable:
    """Wrapper allowing standard functions to be specified as class
    attributes which will not bind as methods.

    This wrapper hides functions' descriptor interface, such that, for
    example, functions intended as ``Enum`` values are not treated as
    methods (and thereby ignored as values).

    """
    def __init__(self, func):
        # Copy over useful, identifying attributes
        # But, disable attributes to copy by "update", which includes
        # __dict__, and thereby copies everything we're trying to hide
        functools.update_wrapper(self, func, updated=())

    def __call__(self, *args, **kwargs):
        return self.__wrapped__(*args, **kwargs)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.__wrapped__!r})"


_c = WrappedCallable


class CallableEnum(enum.Enum):
    """Enum whose instances are callable and which invoke their values."""

    def __call__(self, *args, **kwargs):
        return self.value(*args, **kwargs)


def _noopl(x):
    return x


class CommandDecorator:
    """Decorate a callable to replace it with a manufactured command
    class.

    The callable is supplied as the manufactured command's functionality
    method, (*e.g.* either ``__call__`` or ``prepare``).

    A CLI parser argument may be specified in the initialization phase
    of the decorator instance, and/or through repeated decoration of
    the manufactured command.

    The manufactured command may be further customized through decorator
    initialization keyword flags ``local`` and ``root``, or by explicit
    specification of a custom base class.

    The binding of the decorated callable to the command instance, (and
    therefore the arguments it may expect to receive, such as ``self``),
    is *static* by default. However, if it is a ``Local`` command, the
    callable will instead receive its default binding, (*e.g.* as a
    command instance method if the callable is a standard function). The
    binding may be explicitly controlled by setting the ``binding``
    keyword to either a Boolean value or to a ``Binding`` instance.

    """
    class Binding(CallableEnum):

        default = _c(_noopl)
        static = _c(staticmethod)
        parent = _c(CommandMethod)

    def __init__(self,
                 *parser_args,
                 base=Unset,
                 binding=Unset,
                 local=False,
                 root=False,
                 method_name=None,
                 **parser_kwargs):
        if (local or root) and base is not Unset:
            raise TypeError("cannot apply 'local' or 'root' functionality to "
                            "arbitrary base")

        if base is Unset:
            if local:
                self.base = LocalRoot if root else Local
            else:
                self.base = RootCommand if root else Command
        else:
            self.base = base

        if binding is Unset:
            if issubclass(self.base, Local):
                self.binding = self.Binding.default
            else:
                self.binding = self.Binding.static
        elif isinstance(binding, bool):
            if binding:
                self.binding = self.Binding.default
            else:
                self.binding = self.Binding.static
        elif isinstance(binding, self.Binding):
            self.binding = binding
        else:
            raise TypeError('binding must be either bool, Binding or Unset')

        if method_name is None:
            self.method_name = 'prepare' if issubclass(self.base, Local) else '__call__'
        else:
            self.method_name = method_name

        self.args = (parser_args, parser_kwargs)

    def __call__(self, target):
        args = [self.args] if any(self.args) else []

        if inspect.isclass(target):
            if issubclass(target, GeneratedCommand):
                target._parser_args_.extend(args)
                return target

        elif callable(target):
            return type(
                target.__name__,
                (GeneratedCommand, self.base),
                {
                    '_parser_args_': args,
                    '__doc__': target.__doc__,
                    '__module__': target.__module__,
                    self.method_name: self.binding(target),
                }
            )

        raise TypeError(f"unexpected command decoration target {target}")


def cmd(*args, **kwargs):
    """Decorate a callable to replace it with a manufactured command
    class.

    Extends the interface of ``CommandDecorator``, allowing the same
    ``cmd`` to be used as a decorator or as a decorator factory::

        @cmd(root=True)
        def build():
            ...

        @build.register
        @cmd
        def deploy():
            ...

    Further enables composition of configuration, for example via
    partials, as helpers.

    """
    try:
        (first, *remainder) = args
    except ValueError:
        pass
    else:
        if callable(first):
            return CommandDecorator(*remainder, **kwargs)(first)

    return CommandDecorator(*args, **kwargs)


local = functools.partial(cmd, local=True)

cmdmethod = functools.partial(cmd, binding=CommandDecorator.Binding.parent)

localmethod = functools.partial(local, binding=CommandDecorator.Binding.parent)


def entrypoint(cls):
    """Mark the decorated command as the intended entrypoint of the
    command module.

    """
    if not isinstance(cls, type) or not issubclass(cls, Command):
        raise TypeError(f"inappropriate entrypoint instance of type {cls.__class__}")

    cls._argcmdr_entrypoint_ = True

    return cls


#                   #
# Interface helpers #
#                   #

# TODO: convert to package and move these argparse helpers to submodule


def store_env_override(option_strings,
                       dest,
                       envvar,
                       nargs=None,
                       default=None,
                       type=None,
                       choices=None,
                       description=None,
                       help=None,
                       metavar=None):
    """Construct an argparse action which stores the value of a command
    line option to override a corresponding value in the process
    environment.

    If the environment variable is not empty, then no override is
    required. If the environment variable is empty, and no default is
    provided, then the "option" is required.

    In the case of a default value which is a *transformation* of the
    single environment variable, this default may be provided as a
    callable, (*e.g.* as a lambda function).

    Rather than have to fully explain the relationship of this
    environment-backed option, help text may be generated from a
    provided description.

    """
    if envvar == '':
        raise ValueError("unsupported environment variable name", envvar)

    envvalue = os.getenv(envvar)

    if callable(default):
        default_value = default(envvalue)
    elif envvalue:
        default_value = envvalue
    else:
        default_value = default

    if description and help:
        raise ValueError(
            "only specify help to override its optional generation from "
            "description -- not both"
        )
    elif description:
        if default_value:
            help = '{} (default {} envvar {}: {})'.format(
                description,
                'provided by' if default is None else 'derived from',
                envvar,
                default_value,
            )
        else:
            help = (f'{description} (required because '
                    f'envvar {envvar} is empty)')

    return argparse._StoreAction(
        option_strings=option_strings,
        dest=dest,
        nargs=nargs,
        const=None,
        default=default_value,
        type=type,
        choices=choices,
        required=(not default_value),
        help=help,
        metavar=metavar,
    )



import argparse
import io
import os
import pdb
import re
import subprocess
import types
import unittest
import sys
import tempfile
import traceback
from unittest import mock

import plumbum.commands

from argcmdr import (
    cmd,
    Command,
    CommandMethod,
    execute,
    GeneratedCommand,
    _exhaust_iterable,
    local,
    Local,
    localmethod,
    main,
    RootCommand,
)


ANSI_ESCAPE = re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]')

TEST_DIR = os.path.dirname(__file__)
DATA_DIR = os.path.join(TEST_DIR, 'data')


def pm():
    info = sys.exc_info()
    traceback.print_exception(*info)
    pdb.post_mortem(info[2])


class TryCommandTestCase(unittest.TestCase):

    def try_command(self, command_cls):
        # ensure parser is available to tests with nested command defns
        (self.parser, args) = command_cls._get_parser_()
        self.parser.parse_args([], args)
        command = args._command_
        command.call()


class TryMainTestCase(unittest.TestCase):

    def try_main(self, *args):
        try:
            main(*args, argv=[])
        except SystemExit as exc:
            self.fail(exc)


class TestCommandGetItem(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.Leaf1 = type('Leaf1', (Command,), {})
        cls.Leaf2 = type('Leaf2', (Command,), {})
        cls.Branch1 = type('Branch1', (Command,), {
            'Leaf1': cls.Leaf1,
        })
        cls.Branch2 = type('Branch2', (Command,), {
            'Leaf2': cls.Leaf2,
        })
        cls.Root = type('Root', (Command,), {
            'Branch1': cls.Branch1,
            'Branch2': cls.Branch2,
        })

    def setUp(self):
        interface = self.Root._build_interface_()
        (_parser, _namespace, self.root) = next(interface)
        _exhaust_iterable(interface)

    def test_identity(self):
        self.assertIs(self.root[()], self.root)

    def test_identity_list(self):
        self.assertIs(self.root[[]], self.root)

    def test_descend_1(self):
        self.assertIs(self.root['branch1'].__class__, self.Branch1)

    def test_descend_2(self):
        self.assertIs(self.root['branch1', 'leaf1'].__class__, self.Leaf1)

    def test_descend_2_list(self):
        self.assertIs(self.root[['branch1', 'leaf1']].__class__, self.Leaf1)

    def test_ascend_1(self):
        leaf1 = self.root['branch1', 'leaf1']
        self.assertIs(leaf1.__class__, self.Leaf1)
        self.assertIs(leaf1[-1].__class__, self.Branch1)

    def test_ascend_2(self):
        leaf1 = self.root['branch1', 'leaf1']
        self.assertIs(leaf1.__class__, self.Leaf1)
        self.assertIs(leaf1[-2], self.root)

    def test_strafe(self):
        branch1 = self.root['branch1']
        self.assertIs(branch1[-1, 'branch2'].__class__, self.Branch2)

    def test_missing_key(self):
        with self.assertRaises(LookupError):
            self.root['branch3']

    def test_missing_index(self):
        with self.assertRaises(LookupError):
            self.root[-1]

    def test_bad_key(self):
        with self.assertRaises(TypeError):
            self.root[None]

    def test_bad_index(self):
        with self.assertRaises(ValueError):
            self.root[0]


class TestCommandGetItemError(unittest.TestCase):

    def test_get_item_error(test):
        class BadGetItemCommand(Command):

            def __init__(self, parser):
                super().__init__(parser)

                with test.assertRaises(RuntimeError):
                    self['nope']

            def __call__(self):
                pass

        next(BadGetItemCommand._build_interface_())

    def test_get_item_error_quick(test):
        class QuickBadGetItemCommand(Command):

            def __init__(self, _parser):
                with test.assertRaises(RuntimeError):
                    self['nope']

            def __call__(self):
                pass

        next(QuickBadGetItemCommand._build_interface_())


class TestSubcommandIteration(unittest.TestCase):

    @staticmethod
    def build_interface(command):
        interface = command._build_interface_()
        (_parser, _namespace, root) = next(interface)
        _exhaust_iterable(interface)
        return root

    def test_iter_children(self):
        Leaf1 = type('Leaf1', (Command,), {})
        Leaf2 = type('Leaf2', (Command,), {})
        Branch1 = type('Branch1', (Command,), {
            'Leaf1': Leaf1,
        })
        Branch2 = type('Branch2', (Command,), {
            'Leaf2': Leaf2,
        })
        Root = type('Root', (Command,), {
            'Branch1': Branch1,
            'Branch2': Branch2,
        })

        root = self.build_interface(Root)

        self.assertEqual(list(root), [root['branch1'], root['branch2']])
        self.assertEqual(list(root['branch1']), [root['branch1', 'leaf1']])
        self.assertEqual(list(root['branch2']), [root['branch2', 'leaf2']])
        self.assertEqual(list(root['branch1', 'leaf1']), [])
        self.assertEqual(list(root['branch2', 'leaf2']), [])

    def test_iter_error(test):
        class BadIterCommand(Command):

            def __init__(self, parser):
                super().__init__(parser)

                with test.assertRaises(RuntimeError):
                    list(self)

            def __call__(self):
                pass

        test.build_interface(BadIterCommand)

    def test_iter_error_quick(test):
        class QuickBadIterCommand(Command):

            def __init__(self, _parser):
                with test.assertRaises(RuntimeError):
                    list(self)

            def __call__(self):
                pass

        test.build_interface(QuickBadIterCommand)


class TestCommandDelegation(unittest.TestCase):

    def test_delegation_to_child(test):
        @cmd('--no-eat', action='store_false', default=True, dest='should_eat')
        @cmd(root=True, binding=True)
        def root(self, args):
            test.assertTrue(args.should_eat)
            test.assertFalse(hasattr(args, 'this_food'))
            test.assertIs(args, self.get_args())

            # This would fail!
            # (forcing Child to handle missing 'this_food')
            # self['child'](args)
            #
            # This won't ;)
            self['child'].delegate()

        @root.register
        @cmd('--food', default='snacks', dest='this_food')
        @cmd(binding=True)
        def child(self, args):
            # target command's args populated regardless
            test.assertTrue(args.should_eat)

            # delegation should populate this command's defaults
            test.assertEqual(args.this_food, 'snacks')

            # ...in args property as well
            test.assertEqual(self.args.this_food, 'snacks')

            # self.args is self.delegate_args
            test.assertIs(self.args, self.delegate_args)

            # ...is args
            test.assertIs(args, self.args)

            # ...not the "real" args
            test.assertIsNot(args, self.get_args())

            # ...only the target gets those:
            test.assertIs(self.root.args, self.get_args())

        (parser, args) = root._get_parser_()
        parser.parse_args([], args)
        args._command_.call()

    def test_delegation_to_root(test):
        @cmd('--no-eat', action='store_false', default=True, dest='should_eat')
        @cmd(root=True, binding=True)
        def root(self, args, parser):
            # root *also* has child's args, because child was called
            test.assertTrue(args.should_eat)
            test.assertEqual(args.this_food, 'snacks')

            # root gets delegate_args -- even though it doesn't *strictly* need it
            #
            # (child's args are superset of root's, so this could be omitted in this case,
            # perhaps as an optimization.)
            #
            # however, this way, root gets access to *its* subparser, (on the off-chance that
            # that matters). and, we're just consistent and straight-forward about it.
            test.assertIs(args, self.args)
            test.assertIs(self.args, self.delegate_args)

            test.assertIs(parser, args._parser_)
            test.assertIs(parser, self._parser_)

        @root.register
        @cmd('--food', default='snacks', dest='this_food')
        @cmd(binding=True)
        def child(self, args):
            test.assertTrue(args.should_eat)
            test.assertEqual(args.this_food, 'snacks')

            # self.args is NOT self.delegate_args
            test.assertIs(args, self.get_args())
            test.assertIs(args, self.args)
            test.assertIsNot(self.args, self.delegate_args)

            self.root.delegate()

        (parser, args) = root._get_parser_()
        parser.parse_args(['child'], args)
        args._command_.call()

    def test_delegation_to_sibling(test):
        @cmd('--no-eat', action='store_false', default=True, dest='should_eat', root=True)
        def root():
            raise AssertionError("I should not be invoked")

        @root.register
        @cmd('--candy', default='jelly beans', dest='this_candy', binding=True)
        def left(self, args, parser):
            test.assertTrue(args.should_eat)
            test.assertEqual(args.this_food, 'snacks')

            # delegation should populate this command's defaults
            test.assertEqual(args.this_candy, 'jelly beans')

            # self.args is self.delegate_args
            test.assertIs(self.args, self.delegate_args)

            # ...is args
            test.assertIs(args, self.args)

            # ...not the "real" args
            test.assertIsNot(args, self.get_args())

        @root.register
        @cmd('--food', default='snacks', dest='this_food', binding=True)
        def right(self, args):
            test.assertTrue(args.should_eat)
            test.assertEqual(args.this_food, 'snacks')

            test.assertFalse(hasattr(args, 'this_candy'))

            # self.args is NOT self.delegate_args
            test.assertIs(args, self.get_args())
            test.assertIs(args, self.args)
            test.assertIsNot(self.args, self.delegate_args)

            self[-1, 'left'].delegate()

        (parser, args) = root._get_parser_()
        parser.parse_args(['right'], args)
        args._command_.call()

    def test_arbitrary_access(test):
        @cmd('--no-eat', action='store_false', default=True, dest='should_eat', root=True)
        def root():
            raise AssertionError("I should not be invoked")

        @root.register
        class Left(Command):

            def __init__(self, parser):
                parser.add_argument(
                    '--candy',
                    default='jelly beans',
                    dest='this_candy',
                )

            def __call__(self):
                raise AssertionError("I should not be invoked")

            def method(self):
                test.assertTrue(self.args.should_eat)
                test.assertEqual(self.args.this_food, 'snacks')

                # delegation should populate this command's defaults
                test.assertEqual(self.args.this_candy, 'jelly beans')

                # self.args is self.delegate_args
                test.assertIs(self.args, self.delegate_args)

                # ...not the "real" args
                test.assertIsNot(self.args, self.get_args())

        @root.register
        @cmd('--food', default='snacks', dest='this_food', binding=True)
        def right(self, args):
            test.assertTrue(args.should_eat)
            test.assertEqual(args.this_food, 'snacks')

            test.assertFalse(hasattr(args, 'this_candy'))

            # self.args is NOT self.delegate_args
            test.assertIs(args, self.get_args())
            test.assertIs(args, self.args)
            test.assertIsNot(self.args, self.delegate_args)

            self[-1, 'left'].method()

        (parser, args) = root._get_parser_()
        parser.parse_args(['right'], args)
        args._command_.call()

    def test_parser_defaults(test):
        """delegate_args respects ArgumentParser.set_defaults"""
        class Root(RootCommand):

            def __init__(self, parser):
                parser.add_argument(
                    '--no-eat',
                    action='store_false',
                    default=True,
                    dest='should_eat',
                )

                parser.set_defaults(root='so cool')

            def __call__(self, args):
                test.assertTrue(args.should_eat)
                test.assertEqual(args.root, 'so cool')
                test.assertFalse(hasattr(args, 'this_food'))
                test.assertIs(args, self.get_args())

                self['child'].delegate()

        @Root.register
        class Child(Command):

            def __init__(self, parser):
                parser.add_argument(
                    '--food',
                    default='snacks',
                    dest='this_food',
                )

                parser.set_defaults(child='even cooler')

            def __call__(self, args):
                test.assertTrue(args.should_eat)
                test.assertEqual(args.root, 'so cool')

                # delegation should populate this command's defaults
                test.assertEqual(args.this_food, 'snacks')
                test.assertEqual(args.child, 'even cooler')

                # self.args is self.delegate_args
                test.assertIs(args, self.args)
                test.assertIs(self.args, self.delegate_args)

                # ...not the "real" args
                test.assertIsNot(args, self.get_args())

        (parser, args) = Root._get_parser_()
        parser.parse_args([], args)
        args._command_.call()

    def test_delegation_to_method(test):
        """command may (re)-delegate to named method"""
        @cmd(binding=True, root=True)
        def root(context, args):
            test.assertFalse(hasattr(args, 'argument'))

            context['child'].delegate()

        @root.register
        class Child(Command):

            run_count = 0

            def __init__(self, parser):
                parser.add_argument(
                    '--nope',
                    dest='argument',
                    action='store_false',
                )

            def __call__(self, args):
                test.assertTrue(getattr(args, 'argument', None))

                self.delegate('other')

            def other(self, args):
                self.__class__.run_count += 1

                test.assertTrue(getattr(args, 'argument', None))

        (parser, args) = root._get_parser_()
        parser.parse_args([], args)
        args._command_.call()

        test.assertEqual(Child.run_count, 1)


class TestCommandRoot(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.Leaf = type('Leaf', (Command,), {})
        cls.Branch = type('Branch', (Command,), {
            'Leaf': cls.Leaf,
        })
        cls.Root = type('Root', (Command,), {
            'Branch': cls.Branch,
        })

    def setUp(self):
        for (_parser, _namespace, command) in self.Root._build_interface_():
            setattr(self, command.name, command)

    def test_root_0(self):
        self.assertIsNone(self.root.root)

    def test_root_1(self):
        self.assertIs(self.branch.root, self.root)

    def test_root_2(self):
        self.assertIs(self.leaf.root, self.root)


class TestMainCallSignature(TryMainTestCase):

    def test_simple(self):
        class Simple(Command):

            def __call__(self_, args):
                self.assertIsInstance(self_, Simple)
                self.assertIsInstance(args, argparse.Namespace)

        self.try_main(Simple)

    def test_deep(self):
        class Deep(Command):

            def __call__(self_, args, parser):
                self.assertIsInstance(self_, Deep)
                self.assertIsInstance(args, argparse.Namespace)
                self.assertIsInstance(parser, argparse.ArgumentParser)

        self.try_main(Deep)

    def test_fancy(self):
        class Fancy(Command):

            def __call__(self_, args, parser, lang='en'):
                self.assertIsInstance(self_, Fancy)
                self.assertIsInstance(args, argparse.Namespace)
                self.assertIsInstance(parser, argparse.ArgumentParser)
                self.assertEqual(lang, 'en')

        self.try_main(Fancy)

    def test_lame(self):
        class Lame(Command):

            def __call__(self_):
                self.assertIsInstance(self_, Lame)

        self.try_main(Lame)

    def test_bad(self):
        class Bad(Command):

            def __call__(self_, args, parser, lang):
                self.fail("how did this happen?")

        outfile = io.StringIO()
        with self.assertRaises(SystemExit) as context:
            main(Bad, argv=[], outfile=outfile)

        exc = context.exception.__context__
        self.assertIsInstance(exc, TypeError)
        self.assertEqual(exc.args, ("Bad.__call__() requires too many "
                                    "positional arguments: 'lang'",))
        self.assertIn(exc.args[0], outfile.getvalue())

    def test_tricky(self):
        class Tricky(Command):

            def __call__(self_, args, parser=None, lang='en'):
                self.assertIsInstance(self_, Tricky)
                self.assertIsInstance(args, argparse.Namespace)
                self.assertIsInstance(parser, argparse.ArgumentParser)
                self.assertEqual(lang, 'en')

        self.try_main(Tricky)


class TestPrepareCallSignature(TryCommandTestCase):

    def test_simple(self):
        class Simple(Local):

            def prepare(self_, args):
                self.assertIsInstance(self_, Simple)
                self.assertIsInstance(args, argparse.Namespace)
                self.assertIs(args._command_, self_)
                self.assertIs(args._parser_, self.parser)

        self.try_command(Simple)

    def test_deep(self):
        class Deep(Local):

            def prepare(self_, args, parser):
                self.assertIsInstance(self_, Deep)
                self.assertIsInstance(args, argparse.Namespace)
                self.assertIs(args._command_, self_)
                self.assertIs(args._parser_, self.parser)
                self.assertIs(parser, self.parser)

        self.try_command(Deep)

    def test_local(self):
        class DeepLocal(Local):

            def prepare(self_, args, parser, local):
                self.assertIsInstance(self_, DeepLocal)
                self.assertIsInstance(args, argparse.Namespace)
                self.assertIs(args._command_, self_)
                self.assertIs(args._parser_, self.parser)
                self.assertIs(parser, self.parser)
                self.assertIs(local, self_.local)

        self.try_command(DeepLocal)

    def test_fancy(self):
        class Fancy(Local):

            def prepare(self_, args, parser, local, lang='en'):
                self.assertIsInstance(self_, Fancy)
                self.assertIsInstance(args, argparse.Namespace)
                self.assertIs(args._command_, self_)
                self.assertIs(args._parser_, self.parser)
                self.assertIs(parser, self.parser)
                self.assertEqual(lang, 'en')

        self.try_command(Fancy)

    def test_lame(self):
        class Lame(Local):

            def prepare(self_):
                self.assertIsInstance(self_, Lame)

        self.try_command(Lame)

    def test_bad(self):
        class Bad(Local):

            def prepare(self_, args, parser, local, lang):
                self.fail("how did this happen?")

        with self.assertRaises(TypeError) as context:
            self.try_command(Bad)

        exc = context.exception
        self.assertIsInstance(exc, TypeError)
        self.assertEqual(exc.args, ("Bad.prepare() requires too many "
                                    "positional arguments: 'lang'",))

    def test_tricky(self):
        class Tricky(Local):

            def prepare(self_, args, parser=None, local=None, lang='en'):
                self.assertIsInstance(self_, Tricky)
                self.assertIsInstance(args, argparse.Namespace)
                self.assertIs(args._command_, self_)
                self.assertIs(args._parser_, self.parser)
                self.assertIs(parser, self.parser)
                self.assertEqual(lang, 'en')

        self.try_command(Tricky)


class TestArgsProperty(TryMainTestCase):

    def test_identity(test):
        class ArgsCommand(Command):

            def __call__(self, args):
                test.assertIs(self.args, args)
                test.assertIs(self._args_, args)

        test.try_main(ArgsCommand)

    def test_access_error(test):
        class BadArgsCommand(Command):

            def __init__(self, _parser):
                with test.assertRaises(RuntimeError):
                    self.args

            def __call__(self):
                pass

        test.try_main(BadArgsCommand)

    def test_set_error(test):
        class BadArgsCommand(Command):

            def __init__(self, _parser):
                with test.assertRaises(AttributeError):
                    self.args = None

            def __call__(self):
                pass

        test.try_main(BadArgsCommand)


class TestParserProperty(TryMainTestCase):

    def test_identity(test):
        class ParserCommand(Command):

            def __init__(self, parser):
                super().__init__(parser)
                test.assertIs(self.parser, parser)
                test.assertIs(self._parser_, parser)

            def __call__(self, args, parser):
                test.assertIs(self.parser, parser)
                test.assertIs(self._parser_, parser)

        test.try_main(ParserCommand)

    def test_identity_no_super(test):
        class QuickParserCommand(Command):

            def __init__(self, parser):
                test.assertIs(self.parser, None)
                test.assertFalse(hasattr(self, '_parser_'))
                test.assertIsNot(parser, None)

            def __call__(self, args, parser):
                test.assertIs(self.parser, parser)
                test.assertIs(self._parser_, parser)
                test.assertIsNot(parser, None)

        test.try_main(QuickParserCommand)

    def test_set_error(test):
        class BadParserCommand(Command):

            def __init__(self, _parser):
                with test.assertRaises(AttributeError):
                    self.parser = None

            def __call__(self):
                with test.assertRaises(AttributeError):
                    self.parser = None

        test.try_main(BadParserCommand)


class TestSendCommandResult(TryCommandTestCase):

    def test_execution(self):
        class CarefulCommand(Local):

            def prepare(self_, args):
                # don't clutter test output
                args.foreground = False

                (code, std, err) = yield self_.local['which']['python']

                self.assertEqual(code, 0)
                self.assertTrue(std)
                self.assertFalse(err)

                (code, std, err) = yield self_.local['which']['TOTAL-FAKE']

                self.assertEqual(code, 1)
                self.assertFalse(std)
                self.assertFalse(err)

            prepare.retcode = None

        self.try_command(CarefulCommand)

    def test_dry_run(self):
        class SmartCommand(Local):

            def prepare(self_, args):
                args.execute_commands = False

                (code, std, err) = yield self_.local['which']['python']

                self.assertIsNone(code)
                self.assertIsNone(std)
                self.assertIsNone(err)

                (code, std, err) = yield self_.local['which']['TOTAL-FAKE']

                self.assertIsNone(code)
                self.assertIsNone(std)
                self.assertIsNone(err)

        with mock.patch('sys.stdout', new=io.StringIO()) as output:
            self.try_command(SmartCommand)

        plain_output = ANSI_ESCAPE.sub('', output.getvalue())
        self.assertEqual(
            plain_output,
            '> /usr/bin/which python\n'
            '> /usr/bin/which TOTAL-FAKE\n'
        )

    def test_non_gen(self):
        command = mock.Mock(spec=Local.local['which']['python'])

        class SimpleCommand(Local):

            def prepare(self_, args):
                # mock not configured to handle foreground operation
                args.foreground = False
                return command

        self.try_command(SimpleCommand)
        command.run.assert_called_once_with()


class TestThrowCommandException(TryCommandTestCase):

    def test_throw(self):
        class CarefulCommand(Local):

            def prepare(self_, args):
                # don't clutter test output
                args.foreground = False

                with self.assertRaises(self_.local.ProcessExecutionError):
                    yield self_.local['which']['TOTAL-FAKE']

        self.try_command(CarefulCommand)

    def test_naive(self):
        class NaiveCommand(Local):

            def prepare(self_, args):
                # don't clutter test output
                args.foreground = False

                yield self_.local['which']['TOTAL-FAKE']

        with self.assertRaises(NaiveCommand.local.ProcessExecutionError):
            self.try_command(NaiveCommand)

    def test_non_gen(self):
        class SimpleCommand(Local):

            def prepare(self_, args):
                # don't clutter test output
                args.foreground = False

                return self_.local['which']['TOTAL-FAKE']

        with self.assertRaises(SimpleCommand.local.ProcessExecutionError):
            self.try_command(SimpleCommand)


class TestCommandDecorator(unittest.TestCase):

    @staticmethod
    def command():
        pass

    def test_vanilla_cmd(self):
        command = cmd(self.command)
        self.assertIsNot(command, self.command)
        self.assertIs(command.__call__, self.command)
        self.assertTrue(issubclass(command, GeneratedCommand))
        self.assertTrue(issubclass(command, Command))
        self.assertEqual(command._parser_args_, [])
        self.assertIs(command(None).__call__, self.command)

    def test_parsing_cmd(self):
        parser_args0 = ('-1',)
        parser_kwargs0 = {
            'action': 'store_const',
            'const': '\n',
            'default': ' ',
            'dest': 'sep',
            'help': 'list one file per line',
        }
        parser_args1 = ('-h', '--human-readable')
        parser_kwargs1 = {
            'action': 'store_true',
            'dest': 'human',
            'help': 'print human readable sizes (e.g., 1K 234M 2G)',
        }

        command = cmd(*parser_args0, **parser_kwargs0)(
            cmd(*parser_args1, **parser_kwargs1)(
                self.command
            )
        )
        self.assertIsNot(command, self.command)
        self.assertIs(command.__call__, self.command)
        self.assertTrue(issubclass(command, GeneratedCommand))
        self.assertTrue(issubclass(command, Command))
        self.assertEqual(command._parser_args_, [(parser_args1, parser_kwargs1),
                                                 (parser_args0, parser_kwargs0)])

        parser = mock.Mock()
        instance = command(parser)
        self.assertEqual(parser.add_argument.call_count, 2)
        parser.add_argument.assert_has_calls([(parser_args1, parser_kwargs1),
                                              (parser_args0, parser_kwargs0)])
        self.assertIs(instance.__call__, self.command)

    def test_local_cmd(self):
        command = local(self.command)
        self.assertIsNot(command, self.command)
        self.assertIs(command.prepare, self.command)
        self.assertTrue(issubclass(command, GeneratedCommand))
        self.assertTrue(issubclass(command, Local))
        self.assertEqual(command._parser_args_, [])

        instance = command(None)
        self.assertIsInstance(instance.prepare, types.MethodType)
        self.assertIs(instance.prepare.__func__, self.command)

    def test_local_method(self):
        command = localmethod(self.command)
        self.assertIsNot(command, self.command)
        self.assertIsInstance(command.prepare, CommandMethod)
        self.assertIs(command.prepare.__func__, self.command)
        self.assertTrue(issubclass(command, GeneratedCommand))
        self.assertTrue(issubclass(command, Local))
        self.assertEqual(command._parser_args_, [])

        instance = command(None)
        instance.__getitem__ = mock.Mock()
        self.assertIsInstance(instance.prepare, types.MethodType)
        self.assertIs(instance.prepare.__func__, self.command)
        instance.__getitem__.assert_has_calls(2 * [mock.call((-1,))])


class TestLocalModifier(TryCommandTestCase):

    def test_single_default(self):
        # tough to trick TEE's select.select() with anything but actual
        # file descriptors; so, we'll provide them (as TemporaryFile).

        with tempfile.TemporaryFile() as stdout, \
             tempfile.TemporaryFile() as stderr:

            process = mock.Mock(stdout=stdout,
                                stderr=stderr)
            manager = mock.MagicMock()
            command = mock.Mock(spec=Local.local['which']['python'])
            command.bgrun.return_value = manager
            manager.__enter__.return_value = process

            class SimpleCommand(Local):

                def prepare(_self):
                    return command

            self.try_command(SimpleCommand)

        command.bgrun.assert_called_once_with(
            retcode=0,
            stdin=None,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            timeout=None,
        )
        process.poll.assert_called_once_with()

    def test_single_foreground(self):
        command = mock.Mock(spec=Local.local['which']['python'])

        class SimpleCommand(Local):

            def prepare(self_):
                return (self_.local.FG, command)

        self.try_command(SimpleCommand)
        command.assert_called_once_with(
            retcode=0,
            stderr=None,
            stdin=None,
            stdout=None,
            timeout=None,
        )

    def test_generator_background(self):
        command = mock.Mock(spec=Local.local['which']['python'])

        class SimpleCommand(Local):

            def prepare(self_):
                result = yield (self_.local.BG, command)
                self.assertIsInstance(result, plumbum.commands.Future)
                self.assertIs(result.proc, command.popen.return_value)

        self.try_command(SimpleCommand)
        command.popen.assert_called_once_with()

    def test_simple_shh(self):
        command = mock.Mock(spec=Local.local['which']['python'])

        class SimpleCommand(Local):

            def prepare(self_):
                return (self_.local.SHH, command)

        self.try_command(SimpleCommand)
        command.run.assert_called_once_with(
            retcode=0,
            timeout=None,
        )

    def test_global_params(self):
        command = mock.Mock(spec=Local.local['which']['pythoX'])

        class SimpleCommand(Local):

            def prepare(self_):
                return (self_.local.FG, command)

            prepare.retcode = None

        self.try_command(SimpleCommand)
        command.assert_called_once_with(
            retcode=None,
            stderr=None,
            stdin=None,
            stdout=None,
            timeout=None,
        )

    def test_modifier_params(self):
        command = mock.Mock(spec=Local.local['which']['pythoX'])

        class SimpleCommand(Local):

            def prepare(self_):
                return (self_.local.FG(retcode=None), command)

        self.try_command(SimpleCommand)
        command.assert_called_once_with(
            retcode=None,
            stderr=None,
            stdin=None,
            stdout=None,
            timeout=None,
        )


class TryExecuteTestCase(unittest.TestCase):

    test_target = None

    @classmethod
    def declare_success(cls):
        cls.test_target = 'SUCCESS'

    @classmethod
    def clear_success(cls):
        cls.test_target = None

    def tearDown(self):
        try:
            del sys.modules['manage']
        except KeyError:
            pass

        self.clear_success()

    def try_execute(self, argv=None):
        if argv is None:
            argv = []

        argv.append(self.__class__.__name__)

        try:
            execute(argv=argv)
        except SystemExit as exc:
            self.fail(exc)


class TestExecutePath(TryExecuteTestCase):

    def test_module(self):
        self.try_execute([
            '--manage-file', os.path.join(DATA_DIR, 'execute_1', 'manage.py'),
        ])
        self.assertEqual(self.test_target, 'SUCCESS')

    def test_package(self):
        self.try_execute([
            '--manage-file', os.path.join(DATA_DIR, 'execute_2', 'manage'),
            'subcommand',
        ])
        self.assertEqual(self.test_target, 'SUCCESS')


class TryExecuteCwdTestCase(TryExecuteTestCase):

    sample_path = None

    def setUp(self):
        self.cwd0 = os.getcwd()
        os.chdir(self.sample_path)

        # CWD is usually *not* on PYTHONPATH
        try:
            self.cwd_path_index = sys.path.index('')
        except ValueError:
            self.cwd_on_path = False
            self.cwd_path_index = None
        else:
            self.cwd_on_path = True
            sys.path.remove('')

    def tearDown(self):
        os.chdir(self.cwd0)

        if self.cwd_on_path:
            if sys.path[self.cwd_path_index] != '':
                sys.path.insert(self.cwd_path_index, '')
        elif self.cwd_path_index is not None:
            if sys.path[self.cwd_path_index] == '':
                del sys.path[self.cwd_path_index]

        super().tearDown()


class TestExecuteFile(TryExecuteCwdTestCase):

    sample_path = os.path.join(DATA_DIR, 'execute_1')

    def test_on_pythonpath(self):
        if self.cwd_path_index is None:
            self.cwd_path_index = 0

        sys.path.insert(self.cwd_path_index, '')

        self.try_execute()
        self.assertEqual(self.test_target, 'SUCCESS')

    @unittest.skipIf(sys.version_info >= (3, 7), "inapplicable to python3.7 and higher")
    def test_not_on_pythonpath(self):
        with self.assertRaises(ImportError):
            import manage  # noqa: F401

        self.try_execute()
        self.assertEqual(self.test_target, 'SUCCESS')


class TestExecutePackage(TryExecuteCwdTestCase):

    sample_path = os.path.join(DATA_DIR, 'execute_2')

    def test_on_pythonpath(self):
        if self.cwd_path_index is None:
            self.cwd_path_index = 0

        sys.path.insert(self.cwd_path_index, '')

        self.try_execute(['subcommand'])
        self.assertEqual(self.test_target, 'SUCCESS')

    @unittest.skipIf(sys.version_info >= (3, 7), "inapplicable to python3.7 and higher")
    def test_not_on_pythonpath(self):
        with self.assertRaises(ImportError):
            import manage  # noqa: F401

        self.try_execute(['subcommand'])
        self.assertEqual(self.test_target, 'SUCCESS')


if __name__ == '__main__':
    unittest.main()
